"""
Configures Silabs 5341 clock generator.

On DIOT peripheral board there's embedded FTDI (FT4232) UART-to-I2C bridge.
"""

from pyftdi.ftdi import Ftdi
from pyftdi.i2c import I2cController
from typing import Union, Iterable
from pathlib import Path
import time
import csv


class si5341:
    """Silabs 5341 communication class."""

    def __init__(self, i2c, addr=0x76):
        """Class constructor.

        :param i2c: I2C controller object from pyftdi (after configure())
        :param addr: I2C address
        """
        self.port = i2c.get_port(addr)
        self.gpio = i2c.get_gpio()
        self.__page = None  # device page is unknown without configuration

    def configure(self) -> None:
        """Configure board and enable I2C access."""
        # Buffer enable is connected to BDBUS6
        bufen_mask = 0x1 << 6
        self.gpio.set_direction(bufen_mask, bufen_mask)
        self.gpio.write(bufen_mask)
        # reset page register to known setting
        self.__page = 0
        self.port.write_to(0x1, [self.__page])

    def write(self, addr: int, val: int) -> None:
        """Write to device register."""
        page = addr >> 8
        assert page in [0, 1, 2, 3, 5, 8, 9, 0xA, 0xB]
        # pages #5 and #8 aren't mentioned in the reference manual, but are
        # used by config generated with clockbuilder software

        if page != self.__page:
            self.port.write_to(0x1, [page])
            self.__page = page

        self.port.write_to(addr & 0xFF, [val])

    def read(self, addr: int) -> int:
        """Read from device register."""
        page = addr >> 8
        assert page in [0, 1, 2, 3, 5, 8, 9, 0xA, 0xB]
        # pages #5 and #8 aren't mentioned in the reference manual, but are
        # used by config generated with clockbuilder software

        if page != self.__page:
            self.port.write_to(0x1, [page])
            self.__page = page

        val = self.port.exchange([addr & 0xFF], 1)
        return int.from_bytes(val, "big")

    def nvm_write(self) -> None:
        """Write current configuration to NV memory.

        THIS IS VERY DANGEROUS OPERATION THAT CANNOT BE REVERTED!
        Si5341 devices have two NVM banks that can be programmed only one time.
        """
        # check if there are any free banks available to burn
        bank = self.read(0xe2)
        if bank == 3:
            print('Two banks left')
        elif bank == 15:
            print('One bank left')
        elif bank == 63:
            raise Exception('No free NVM banks available to burn')
        else:
            raise Exception(f"Unknown status ({bank}), can't proceed")

        answer = input('Do you want to proceed? [yes]/[no]')
        if any(answer.lower() == f for f in ['yes', 'y']):
            pass
        else:
            print('Aborting NVM burn')
            return

        print('Starting NVM burn procedure')
        self.write(0xe2, 0xc7)
        while self.read(0xfe) != 0x0F:
            pass

        print('Burning finished, checking NVM download')
        self.write(0xe4, 0x1)
        while self.read(0xfe) != 0x0F:  # wait until download is finished
            pass
        new_bank = self.read(0xe2)
        if (bank == 3 and new_bank == 15) or (bank == 15 and new_bank == 63):
            print('NVM programmed properly')

    def load_from_csv(self, path: Union[str, Path]) -> None:
        """Send device configuration from CSV file generated by ClockBuilder.

        :param path: Path to csv file
        """
        def decomment(csvfile):
            for row in csvfile:
                raw = row.split('#')[0].strip()
                if raw:
                    yield raw

        with open(path) as csvfile:
            reader = csv.reader(decomment(csvfile))
            # skip header text
            next(reader)
            ldata = list(reader)
            for idx, row in enumerate(ldata):
                addr = int(row[0], base=16)
                val = int(row[1], base=16)
                self.write(addr, val)
                if addr == 0x0B4E:
                    # Comment in the CSV file suggests waiting after preamble
                    time.sleep(0.3)

            print('Config sent, starting verification readback')
            # skip preamble and postamble
            for idx, row in enumerate(ldata[2:-3]):
                addr = int(row[0], base=16)
                val = int(row[1], base=16)
                rval = self.read(addr)
                if val != rval:
                    print(f'Data mismatch - addr={hex(addr)} sent={hex(val)} '
                          f'received={hex(rval)}')

        time.sleep(1)
        val = self.read(0xC)
        print('Done')
        print(f'Device status (should be 0xC0): {hex(val)}')


print('Connecting to USB-I2C bridge')
i2c = I2cController()
i2c.configure('ftdi://ftdi:4232h/2')
dev = si5341(i2c, 0x76)

print('Enabling USB access to I2C bus')
dev.configure()
fname = 'Si5341-RevD-CCE-Registers.txt'
print(f'Loading configuration data from {fname}')
data = dev.load_from_csv(fname)
